# -*- coding: utf-8 -*-
"""Factor risk model
"""
from __future__ import annotations

from dataclasses import dataclass

import cvxpy as cvx
import numpy as np

from cvx.risk.linalg import cholesky
from cvx.risk.model import RiskModel


@dataclass
class FactorModel(RiskModel):
    """Factor risk model"""

    assets: int = 0
    k: int = 0

    def __post_init__(self):
        self.exposure = cvx.Parameter(
            shape=(self.k, self.assets),
            name="exposure",
            value=np.zeros((self.k, self.assets)),
        )

        self.idiosyncratic_risk = cvx.Parameter(
            shape=self.assets, name="idiosyncratic risk", value=np.zeros(self.assets)
        )

        self.chol = cvx.Parameter(
            shape=(self.k, self.k),
            name="cholesky of covariance",
            value=np.zeros((self.k, self.k)),
        )

        self.lower = cvx.Parameter(
            shape=self.assets,
            name="lower bound",
            value=np.zeros(self.assets),
        )
        self.upper = cvx.Parameter(
            shape=self.assets,
            name="upper bound",
            value=np.ones(self.assets),
        )

    def estimate_risk(self, weights, **kwargs):
        """
        Compute the total variance
        """
        var_residual = cvx.norm2(cvx.multiply(self.idiosyncratic_risk, weights))

        y = kwargs.get("y", self.exposure @ weights)

        return cvx.norm2(cvx.vstack([cvx.norm2(self.chol @ y), var_residual]))

    def update_data(self, **kwargs):
        exposure = kwargs["exposure"]
        k, assets = exposure.shape

        self.exposure.value[:k, :assets] = kwargs["exposure"]
        self.idiosyncratic_risk.value[:assets] = kwargs["idiosyncratic_risk"]
        self.chol.value[:k, :k] = cholesky(kwargs["cov"])

        self.lower.value = np.zeros(self.assets)
        self.lower.value[:assets] = kwargs["lower"]

        self.upper.value = np.zeros(self.assets)
        self.upper.value[:assets] = kwargs["upper"]



# -*- coding: utf-8 -*-
"""Abstract risk model
"""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass


@dataclass
class RiskModel(ABC):
    """Abstract risk model"""

    @abstractmethod
    def estimate_risk(self, weights, **kwargs):
        """
        Estimate the variance given the portfolio weights
        """

    def update_data(self, **kwargs):
        """
        Update the data in the risk model
        """



# -*- coding: utf-8 -*-
"""Risk models based on the sample covariance matrix
"""
from __future__ import annotations

from dataclasses import dataclass

import cvxpy as cvx
import numpy as np

from cvx.risk.linalg import cholesky
from cvx.risk.model import RiskModel
from cvx.risk.random import rand_cov


@dataclass
class SampleCovariance(RiskModel):
    """Risk model based on the Cholesky decomposition of the sample cov matrix"""

    num: int = 0

    def __post_init__(self):
        self._chol = cvx.Parameter(
            shape=(self.num, self.num),
            name="cholesky of covariance",
            value=np.zeros((self.num, self.num)),
        )
        self.lower = cvx.Parameter(
            shape=self.num,
            name="lower bound",
            value=np.zeros(self.num),
        )
        self.upper = cvx.Parameter(
            shape=self.num,
            name="upper bound",
            value=np.ones(self.num),
        )

    def estimate_risk(self, weights, **kwargs):
        """Estimate the risk by computing the Cholesky decomposition of self.cov"""
        return cvx.norm2(self._chol @ weights)

    def update_data(self, **kwargs):
        cov = kwargs["cov"]
        n = cov.shape[0]

        self._chol.value[:n, :n] = cholesky(cov)
        self.lower.value = np.zeros(self.num)
        self.lower.value[:n] = kwargs["lower"]

        self.upper.value = np.zeros(self.num)
        self.upper.value[:n] = kwargs["upper"]


if __name__ == "__main__":
    s = SampleCovariance(num=5)
    cov = rand_cov(4)
    s.update_data(cov=cov, lower=np.zeros(4), upper=np.ones(4))

    print(s.lower.value)
    print(s.upper.value)

    # z = np.zeros((5,5))
    # z[:4,:4] = cholesky(cov)
    # print(z



# -*- coding: utf-8 -*-
from __future__ import annotations

import numpy as np


def valid(matrix):
    """
    Construct the valid subset of matrix (correlation) matrix matrix
    :param matrix: n x n matrix

    :return: Tuple of matrix boolean vector indicating if row/column
    is valid and the valid subset of the matrix
    """
    # make sure matrix  is quadratic
    if matrix.shape[0] != matrix.shape[1]:
        raise AssertionError

    v = np.isfinite(np.diag(matrix))
    return v, matrix[:, v][v]



# -*- coding: utf-8 -*-
from __future__ import annotations

import scipy as sc


def cholesky(cov):
    # upper triangular part of the cholesky decomposition
    # np.linalg.cholesky(cov) is the lower triangular part
    return sc.linalg.cholesky(cov)



# -*- coding: utf-8 -*-
from __future__ import annotations

from dataclasses import dataclass

import cvxpy as cvx
import numpy as np

from cvx.risk.model import RiskModel


@dataclass
class CVar(RiskModel):
    """Conditional value at risk model"""

    alpha: float = 0.95
    n: int = 0
    m: int = 0

    def __post_init__(self):
        self.k = int(self.n * (1 - self.alpha))
        self.R = cvx.Parameter(
            shape=(self.n, self.m), name="returns", value=np.zeros((self.n, self.m))
        )
        self.lower = cvx.Parameter(
            shape=self.m,
            name="lower bound",
            value=np.zeros(self.m),
        )
        self.upper = cvx.Parameter(
            shape=self.m,
            name="upper bound",
            value=np.ones(self.m),
        )

    # R: np.ndarray = None

    def estimate_risk(self, weights, **kwargs):
        """Estimate the risk by computing the Cholesky decomposition of self.cov"""
        # R is a matrix of returns, n is the number of rows in R
        # n = self.R.shape[0]
        # k is the number of returns in the left tail
        # k = int(n * (1 - self.alpha))
        # average value of the k elements in the left tail
        return -cvx.sum_smallest(self.R @ weights, k=self.k) / self.k

    def update_data(self, **kwargs):
        ret = kwargs["returns"]
        m = ret.shape[1]

        self.R.value[:, :m] = kwargs["returns"]

        self.lower.value = np.zeros(self.m)
        self.lower.value[:m] = kwargs["lower"]

        self.upper.value = np.zeros(self.m)
        self.upper.value[:m] = kwargs["upper"]



# -*- coding: utf-8 -*-
"""PCA analysis
"""
from __future__ import annotations

from collections import namedtuple

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA as sklearnPCA

PCA = namedtuple(
    "PCA",
    ["explained_variance", "factors", "exposure", "cov", "systematic", "idiosyncratic"],
)


def pca(returns, n_components=10):
    """
    Compute the first n principal components for a return matrix

    Args:
        returns: DataFrame of prices
        n_components: Number of components
    """

    # USING SKLEARN. Let's look at the first n components
    sklearn_pca = sklearnPCA(n_components=n_components)
    sklearn_pca.fit_transform(returns)

    exposure = sklearn_pca.components_
    factors = returns @ np.transpose(exposure)

    return PCA(
        explained_variance=sklearn_pca.explained_variance_ratio_,
        factors=factors,
        exposure=pd.DataFrame(data=exposure, columns=returns.columns),
        cov=factors.cov(),
        systematic=pd.DataFrame(
            data=factors.values @ exposure, index=returns.index, columns=returns.columns
        ),
        idiosyncratic=pd.DataFrame(
            data=returns.values - factors.values @ exposure,
            index=returns.index,
            columns=returns.columns,
        ),
    )
